let rec string_of_url_path' = function
  | [] -> ""
  | [a] when a = Eliom_common.eliom_suffix_internal_name -> ""
  | [a] -> Eliom_pervasives.Url.encode ~plus:false a
  | a::b::l when b = Eliom_common.eliom_suffix_internal_name -> 
      string_of_url_path' (a::l)
  | a::l when a = Eliom_common.eliom_suffix_internal_name ->
      string_of_url_path' l
  | a::l -> (Eliom_pervasives.Url.encode ~plus:false a)^"/"^(string_of_url_path' l)

let rec string_of_url_path_suff u = function
  | None -> string_of_url_path' u
  | Some suff -> 
      let pref = string_of_url_path' u in
      let suf = string_of_url_path' suff in
      if pref = ""
      then suf
      else String.concat "/" [pref; suf]

let reconstruct_absolute_url_path = string_of_url_path_suff

let reconstruct_relative_url_path current_url u =
  let rec drop cururl desturl = match cururl, desturl with
  | _::l, [_] -> l, desturl
  | [_], m -> [], m
  | a::l, b::m when a = b -> drop l m
  | _::l, m -> l, m
  | [], m -> [], m
  in let rec makedotdot = function
    | [] -> []
(*    | [a] -> "" *)
    | _::l -> ".."::(makedotdot l)
  in
  let aremonter, aaller = drop current_url u in 
  (makedotdot aremonter)@aaller

let reconstruct_relative_url_path_string current_url u suff =
  let relurl = reconstruct_relative_url_path current_url u in
  let s = string_of_url_path_suff relurl suff in
  if s = "" 
  then Eliom_common.defaultpagename 
  else if s.[0] = '/'
  then (* possible with optional parameters *) "./"^s
  else s

let rec relative_url_path_to_myself = function
  | []
  | [""] -> Eliom_common.defaultpagename
  | [a] -> a
  | _::l -> relative_url_path_to_myself l

let make_proto_prefix
    ?hostname
    ?port
    ~sp
    https
    : string =
  let ssl =
    match sp with
      | Some sp -> Eliom_request_info.get_csp_ssl_sp sp
      | None -> false
  in
  let host = match hostname, sp with
    | None, Some sp -> Eliom_request_info.get_csp_hostname_sp sp
    | None, None -> Eliom_config.get_default_hostname () 
    | Some h, _ -> h
  in
  let port = 
    match port, sp with
      | Some p, _ -> p
      | None, Some sp ->
          if https = ssl
          then Eliom_request_info.get_csp_server_port_sp sp
          else if https
          then Eliom_config.get_default_sslport ()
          else Eliom_config.get_default_port ()
      | None, None ->
        if https
        then Eliom_config.get_default_sslport ()
        else Eliom_config.get_default_port ()
  in
  Eliom_pervasives.Url.make_absolute_url https host port "/"


let make_uri_components_ (* does not take into account getparams *)
    ?(absolute = false) (* used to force absolute link *)
    ?(absolute_path = false) (* used to force absolute link without protocol/server/port *)
    ?https
    ~service
    ?hostname
    ?port
    ?fragment
    ?keep_nl_params
    ?(nl_params = Eliom_parameters.empty_nl_params_set) 
    () =

  let sp = Eliom_common.get_sp_option () in
  let ssl =
    match sp with
      | Some sp -> Eliom_request_info.get_csp_ssl_sp sp
      | None -> false
  in
  let https = 
    (https = Some true) || 
      (Eliom_services.get_https service) ||
      (https = None && ssl)
  in
  let absolute = 
    if absolute || https <> ssl 
    then Some (make_proto_prefix ~sp ?hostname ?port https)
    else if absolute_path
    then Some "/"
    else None 
  in
(*VVV We trust current protocol? *) 

  let nl_params = Eliom_parameters.table_of_nl_params_set nl_params in
  let keep_nl_params = match keep_nl_params with
    | None -> Eliom_services.keep_nl_params service
    | Some b -> b
  in
  (* for preapplied non localized and not non localized: *)
  let preappnlp, preapplied_params = Eliom_services.get_pre_applied_parameters_ service in
  let nlp =
    match keep_nl_params, sp with
      | `All, Some sp ->
          (* We replace current nl params by preapplied ones *)
          Eliom_pervasives.String.Table.fold
            (fun key v b -> Eliom_pervasives.String.Table.add key v b)
            preappnlp
            (Eliom_request_info.get_nl_get_params_sp sp)
      | `Persistent, Some sp ->
          (* We replace current nl params by preapplied ones *)
          Eliom_pervasives.String.Table.fold
            (fun key v b -> Eliom_pervasives.String.Table.add key v b)
            preappnlp
            (Eliom_request_info.get_persistent_nl_get_params_sp sp)
      | `All, None
      | `Persistent, None
      | `None, _ -> preappnlp
  in
  let nlp =
    (* We replace current nl params by nl_params *)
    Eliom_pervasives.String.Table.fold
      (fun key v b -> Eliom_pervasives.String.Table.add key v b)
      nl_params
      nlp
  in

  (* remove in nlp the one present in the service parameters *)
  let getparamstype = Eliom_services.get_get_params_type_ service in
  let nlp = Eliom_parameters.remove_from_nlp nlp getparamstype in
  let hiddenparams = 
    Eliom_pervasives.String.Table.fold
      (fun _ l beg -> l@beg)
      nlp preapplied_params
  in

  match Eliom_services.get_kind_ service with
    | `Attached attser ->
        begin
          let uri =
            let suff= None in
            if (Eliom_services.get_att_kind_ attser) = `External
            then
              (Eliom_services.get_prefix_ attser)^
                "/"^  (* we add the "/" even if there is no prefix,
                         because we should do absolute links in that case *)
                (reconstruct_absolute_url_path (Eliom_services.get_full_path_ attser) suff)
            else
              match absolute, sp with
                | Some proto_prefix, _ ->
                    proto_prefix^
                      reconstruct_absolute_url_path
                      (Eliom_services.get_full_path_ attser) suff
                | None, Some sp ->
                    reconstruct_relative_url_path_string
                      (Eliom_request_info.get_csp_original_full_path_sp sp)
                      (Eliom_services.get_full_path_ attser) suff
                | None, None ->
                    reconstruct_relative_url_path_string
                      []
                      (Eliom_services.get_full_path_ attser) suff
          in
          match Eliom_services.get_get_name_ attser, sp with
            | Eliom_common.SAtt_no, _ ->
                (uri, hiddenparams, fragment)
            | Eliom_common.SAtt_anon s, _ ->
                (uri, 
                 ((Eliom_common.get_numstate_param_name, s)::hiddenparams),
                 fragment)
            | Eliom_common.SAtt_named s, _ ->
                (uri,
                 ((Eliom_common.get_state_param_name, s)::hiddenparams),
                 fragment)
            | Eliom_common.SAtt_csrf_safe csrf_info, Some sp ->
                let s = 
                  Eliom_services.register_delayed_get_or_na_coservice
                    ~sp csrf_info
                in
                (uri, 
                 ((Eliom_common.get_numstate_param_name, s)::hiddenparams),
                 fragment)
            | Eliom_common.SAtt_csrf_safe csrf_info, None ->
              failwith "make_uri_component: not possible on csrf safe service not during a request"

        end
    | `Nonattached naser ->
        let na_name = Eliom_services.get_na_name_ naser in
        let params' =
          let current_get_params =
            match sp with
              | None -> []
              | Some sp ->
                (if na_name = Eliom_common.SNa_void_keep
                 then (Eliom_request_info.get_si sp).Eliom_common.si_all_get_but_nl
                 else Lazy.force 
                    (Eliom_request_info.get_si sp).Eliom_common.si_all_get_but_na_nl)
          in
          (match na_name, sp with
             | Eliom_common.SNa_void_keep, _
             | Eliom_common.SNa_void_dontkeep, _ -> current_get_params
             | Eliom_common.SNa_get' n, _ ->
                 (Eliom_common.naservice_num, n)::current_get_params
             | Eliom_common.SNa_get_ n, _ ->
                 (Eliom_common.naservice_name, n)::current_get_params
             | Eliom_common.SNa_get_csrf_safe csrf_info, Some sp ->
                 let n = 
                   Eliom_services.register_delayed_get_or_na_coservice ~sp
                     csrf_info
                 in
                 (Eliom_common.naservice_num, n)::current_get_params
             | Eliom_common.SNa_get_csrf_safe csrf_info, None ->
               failwith "make_uri_component: not possible on csrf safe service outside request"
             | _ -> assert false)
        in
        let params = params'@hiddenparams in
        let beg =
          match absolute, sp with
            | Some proto_prefix, Some sp ->
              proto_prefix^ Eliom_request_info.get_original_full_path_string_sp sp
            | None, Some sp -> 
              relative_url_path_to_myself
                (Eliom_request_info.get_csp_original_full_path_sp sp)
            | Some proto_prefix, None ->
              proto_prefix
            | None, None -> 
              relative_url_path_to_myself []
        in
        (beg, params, fragment)

let make_post_uri_components_ (* do not take into account postparams *)
    ?(absolute = false)
    ?(absolute_path = false)
    ?https
    ~service
    ?hostname
    ?port
    ?fragment
    ?keep_nl_params
    ?(nl_params = Eliom_parameters.empty_nl_params_set) 
    ?(keep_nl_params : [ `All | `Persistent | `None ] option)
    ?keep_get_na_params
    getparams 
    () =
  let sp = Eliom_common.get_sp () in
  match Eliom_services.get_kind_ service with
    | `Attached attser ->
      let (uri, getparams, fragment), getname =
        let getname = Eliom_services.get_get_name_ attser in
        match getname with
          | Eliom_common.SAtt_csrf_safe csrf_info ->
            (* special case for post-coservices on get csrf safe services:
               we must register the get service first *)
            let s = 
              Eliom_common.SAtt_anon 
                (Eliom_services.register_delayed_get_or_na_coservice
                   ~sp csrf_info)
            in
            (Eliom_uri.make_uri_components
               ~absolute
               ~absolute_path
               ?https
               ~service:(Eliom_services.change_get_num service attser s)
               ?hostname
               ?port
               ?fragment
               ?keep_nl_params
               ~nl_params
               getparams,
             s)
          | _ -> (Eliom_uri.make_uri_components
                    ~absolute
                    ~absolute_path
                    ?https
                    ~service
                    ?hostname
                    ?port
                    ?fragment
                    ?keep_nl_params
                    ~nl_params
                    getparams,
                  getname)
      in
      let postparams =
        match Eliom_services.get_post_name_ attser with
          | Eliom_common.SAtt_no -> []
          | Eliom_common.SAtt_anon s ->
            [(Eliom_common.post_numstate_param_name, s)]
          | Eliom_common.SAtt_named s ->
            [(Eliom_common.post_state_param_name, s)]
          | Eliom_common.SAtt_csrf_safe csrf_info ->
            let s =
              Eliom_services.register_delayed_post_coservice
                ~sp csrf_info getname
            in
            [(Eliom_common.post_numstate_param_name, s)]
      in
      (uri, getparams, fragment, postparams)
    | `Nonattached naser ->
      let nl_params = Eliom_parameters.table_of_nl_params_set nl_params in
      let keep_nl_params = match keep_nl_params with
        | None -> Eliom_services.keep_nl_params service
        | Some b -> b
      in
      let preappnlp, preapp = Eliom_services.get_pre_applied_parameters_ service in
      let nlp =
        match keep_nl_params with
          | `All ->
            (* We replace current nl params by preapplied ones *)
            Eliom_pervasives.String.Table.fold
              (fun key v b -> Eliom_pervasives.String.Table.add key v b)
              preappnlp
              (Eliom_request_info.get_nl_get_params ())
          | `Persistent ->
            (* We replace current nl params by preapplied ones *)
            Eliom_pervasives.String.Table.fold
              (fun key v b -> Eliom_pervasives.String.Table.add key v b)
              preappnlp
              (Eliom_request_info.get_persistent_nl_get_params_sp sp)
          | `None -> preappnlp
      in
      let nlp =
        (* We replace current nl params by nl_params *)
        Eliom_pervasives.String.Table.fold
          (fun key v b -> Eliom_pervasives.String.Table.add key v b)
          nl_params
          nlp
      in
      (* for getparams and non localized params: *)
      let suff, params =
        Eliom_parameters.construct_params_list
          nlp (Eliom_services.get_get_params_type_ service) getparams 
            (* if nl params were already present, they will be replaced
               by new values *)
            (* getparams can be something else than [] 
               if we have added nl params to the service (?) *)
      in
      let params = params @ preapp in
      let keep_get_na_params =
        match keep_get_na_params with
          | Some b -> b
          | None ->
            match Eliom_services.get_na_kind_ naser with
              | `Post b -> b
              | _ -> assert false
      in
      let params =
        params @
          (if keep_get_na_params
           then
              (Eliom_request_info.get_si sp).Eliom_common.si_all_get_but_nl
           else
              (Lazy.force
                 (Eliom_request_info.get_si sp).Eliom_common.si_all_get_but_na_nl))
      in
      let ssl = Eliom_request_info.get_csp_ssl_sp sp in
      let https = 
        (https = Some true) || 
          (Eliom_services.get_https service) ||
          (https = None && ssl)
      in
      let absolute = 
        if absolute || https <> ssl 
        then Some (make_proto_prefix ~sp:(Some sp) ?hostname ?port https)
        else if absolute_path
        then Some "/"
        else None
      in
            (* absolute URL does not work behind a reverse proxy! *)
      let uri =
        match absolute with
          | Some proto_prefix ->
            proto_prefix^Eliom_request_info.get_original_full_path_string_sp sp
          | None ->
            relative_url_path_to_myself
              (Eliom_request_info.get_csp_original_full_path_sp sp)
      in
      let naservice_line =
        match Eliom_services.get_na_name_ naser with
          | Eliom_common.SNa_post' n -> (Eliom_common.naservice_num, n)
          | Eliom_common.SNa_post_ n -> (Eliom_common.naservice_name, n)
          | Eliom_common.SNa_post_csrf_safe csrf_info ->
            let n =
              Eliom_services.register_delayed_get_or_na_coservice
                ~sp csrf_info
            in
            (Eliom_common.naservice_num, n)
          | _ -> assert false
      in
      let fragment = None (* fragment is not sent to the server *) in
      let postparams = [naservice_line] in
      (uri, params, fragment, postparams)
	
